@@ -1,3 +1,5 @@
+import os
+import tempfile
 import numpy as np
 import pandas as pd
 import scipy.signal as signal
@@ -8,7 +10,7 @@ import warnings
 __all__ = ['lowpass', 'calibrate_gravity', 'detect_nonwear', 'resample', 'get_stationary_indicator']
 
 
-def resample(data, sample_rate, dropna=False):
+def resample(data, sample_rate, dropna=False, chunksize=1_000_000):
     """
     Nearest neighbor resampling. For downsampling, it is recommended to first
     apply an antialiasing filter.
@@ -19,6 +21,8 @@ def resample(data, sample_rate, dropna=False):
     :type sample_rate: int or float
     :param dropna: Whether to drop NaN values after resampling. Defaults to False.
     :type dropna: bool, optional
+    :param chunksize: Chunk size. Defaults to 1_000_000.
+    :type chunksize: int, optional
     :return: Processed data and processing info.
     :rtype: (pandas.DataFrame, dict)
     """
@@ -35,13 +39,49 @@ def resample(data, sample_rate, dropna=False):
     info['ResampleRate'] = sample_rate
 
     t0, tf = data.index[0], data.index[-1]
-    nt = int(np.around((tf - t0).total_seconds() * sample_rate))  # integer number of ticks we need
-    tf = t0 + pd.Timedelta(nt / sample_rate, unit='s')  # adjust end tick
-    t = pd.date_range(t0, tf, periods=nt + 1, name='time')
+    nt = int(np.around((tf - t0).total_seconds() * sample_rate)) + 1  # integer number of ticks we need
 
-    data = data.reindex(t, method='nearest', tolerance=pd.Timedelta('1s'), limit=1)
+    # # In-memory version
+    # tf = t0 + pd.Timedelta((nt - 1) / sample_rate, unit='s')  # adjust tf
+    # t = pd.date_range(t0, tf, periods=nt, name=data.index.name)
+    # data = data.reindex(t, method='nearest', tolerance=pd.Timedelta('1s'), limit=1)
+
+    with tempfile.TemporaryDirectory() as tmpdir:
+        # We use TemporaryDirectory() + filename instead of NamedTemporaryFile()
+        # because we don't want to open the file just yet:
+        # https://stackoverflow.com/questions/26541416/generate-temporary-file-names-without-creating-actual-file-in-python
+        # and Windows doesn't allow opening a file twice:
+        # https://docs.python.org/3.9/library/tempfile.html#tempfile.NamedTemporaryFile
+        mmap_fname = os.path.join(tmpdir, 'data.mmap')
+
+        data_mmap = mmap_like(data, mmap_fname, shape=(nt,))
+
+        for i in range(0, nt, chunksize):
+
+            # If last chunk, adjust chunksize
+            if i + chunksize > nt:
+                chunksize = nt - i
+
+            # Use pd.Timedelta(n/r) instead of n * pd.Timedelta(1/r): it's not the same due to numerical precision
+            t = pd.date_range(
+                t0 + pd.Timedelta(i / sample_rate, unit='s'),
+                t0 + pd.Timedelta((i + chunksize - 1) / sample_rate, unit='s'),
+                periods=chunksize,
+                name=data.index.name,
+            )
+            chunk = data.reindex(t, method='nearest', tolerance=pd.Timedelta('1s'), limit=1)
+            copy2mmap(chunk, data_mmap[i:i + chunksize])
+
+        del data
+
+        # We need to copy so that the mmap file can be trully deleted: 
+        # https://stackoverflow.com/questions/24178460/in-python-is-it-possible-to-overload-numpys-memmap-to-delete-itself-when-the-m
+        data = mmap2df(data_mmap, copy=True)
+
+        del data_mmap
 
     if dropna:
+        # TODO: This may force a copy of the data
         data = data.dropna()
 
     info['NumTicksAfterResample'] = len(data)
@@ -49,7 +89,7 @@ def resample(data, sample_rate, dropna=False):
     return data, info
 
 
-def lowpass(data, data_sample_rate, cutoff_rate=20):
+def lowpass(data, data_sample_rate, cutoff_rate=20, chunksize=1_000_000):
     """
     Apply Butterworth low-pass filter.
 
@@ -59,42 +99,73 @@ def lowpass(data, data_sample_rate, cutoff_rate=20):
     :type data_sample_rate: int or float
     :param cutoff_rate: Cutoff (Hz) for low-pass filter. Defaults to 20.
     :type cutoff_rate: int, optional
+    :param chunksize: Chunk size. Defaults to 1_000_000.
+    :type chunksize: int, optional
     :return: Processed data and processing info.
     :rtype: (pandas.DataFrame, dict)
     """
 
     info = {}
 
-    orig_index = data.index
-    data, _ = resample(data, data_sample_rate, dropna=False)
-
-    # Butter filter to remove high freq noise.
-    # Default: 20Hz (most of human motion is under 20Hz)
     # Skip this if the Nyquist freq is too low
-    if data_sample_rate / 2 > cutoff_rate:
-        xyz = data[['x', 'y', 'z']].to_numpy()
-        # Temporarily replace nans with 0s for butterfilt
-        where_nan = np.isnan(xyz).any(1)
-        xyz[where_nan] = 0
-        xyz = butterfilt(xyz, cutoff_rate, fs=data_sample_rate, axis=0)
-        # Now restore nans
-        xyz[where_nan] = np.nan
-        data[['x', 'y', 'z']] = xyz
-        info['LowpassOK'] = 1
-        info['LowpassCutoff(Hz)'] = cutoff_rate
-    else:
+    if data_sample_rate / 2 <= cutoff_rate:
         print(f"Skipping lowpass filter: data sample rate {data_sample_rate} too low for cutoff rate {cutoff_rate}")
         info['LowpassOK'] = 0
+        return data, info
 
-    data = data.reindex(orig_index,
-                        method='nearest',
-                        tolerance=pd.Timedelta('1s'),
-                        limit=1)
+    # # In-memory version
+    # xyz = data[['x', 'y', 'z']].to_numpy()
+    # where_nan = np.isnan(xyz).any(1)  # temporarily replace nans with 0s for butterfilt
+    # xyz[where_nan] = 0
+    # xyz = butterfilt(xyz, cutoff_rate, fs=data_sample_rate, axis=0)
+    # xyz[where_nan] = np.nan  # restore nans
+    # data = data.copy(deep=True)  # copy to avoid modifying original data
+    # data[['x', 'y', 'z']] = xyz
+
+    with tempfile.TemporaryDirectory() as tmpdir:
+        # We use TemporaryDirectory() + filename instead of NamedTemporaryFile()
+        # because we don't want to open the file just yet:
+        # https://stackoverflow.com/questions/26541416/generate-temporary-file-names-without-creating-actual-file-in-python
+        # and Windows doesn't allow opening a file twice:
+        # https://docs.python.org/3.9/library/tempfile.html#tempfile.NamedTemporaryFile
+        mmap_fname = os.path.join(tmpdir, 'data.mmap')
+
+        n = len(data)
+        data_mmap = mmap_like(data, mmap_fname, shape=(n,))
+
+        leeway = 100  # used to minimize edge effects
+        for i in range(0, n, chunksize):
+
+            leeway0 = min(i, leeway)
+            istart = i - leeway0
+            istop = i + chunksize + leeway
+
+            chunk = data.iloc[istart : istop]
+            xyz = chunk[['x', 'y', 'z']].to_numpy()
+            na = np.isnan(xyz).any(1)
+            xyz[na] = 0.0  # temporarily replace nans with 0s for butterfilt
+            xyz = butterfilt(xyz, cutoff_rate, fs=data_sample_rate, axis=0)
+            xyz[na] = np.nan  # restore nans
+            chunk = chunk.copy(deep=True)  # copy to avoid modifying original data
+            chunk[['x', 'y', 'z']] = xyz
+            chunk = chunk.iloc[leeway0 : leeway0 + chunksize]  # trim leeway
+            copy2mmap(chunk, data_mmap[i:i + chunksize])
+
+        del data
+
+        # We need to copy so that the mmap file can be trully deleted: 
+        # https://stackoverflow.com/questions/24178460/in-python-is-it-possible-to-overload-numpys-memmap-to-delete-itself-when-the-m
+        data = mmap2df(data_mmap, copy=True)
+
+        del data_mmap
+
+    info['LowpassOK'] = 1
+    info['LowpassCutoff(Hz)'] = cutoff_rate
 
     return data, info
 
 
-def detect_nonwear(data, patience='90m', stationary_indicator=None, drop=False):
+def detect_nonwear(data, patience='90m', window='10s', stdtol=15 / 1000):
     """
     Detect nonwear episodes based on long periods of no movement.
 
@@ -116,36 +187,53 @@ def detect_nonwear(data, patience='90m', stationary_indicator=None, drop=False):
 
     info = {}
 
-    if stationary_indicator is None:
-        stationary_indicator = get_stationary_indicator(data)
+    stationary_indicator = (  # this is more memory friendly than data[['x', 'y', 'z']].std()
+        data['x'].resample(window, origin='start').std().lt(stdtol)
+        & data['y'].resample(window, origin='start').std().lt(stdtol)
+        & data['z'].resample(window, origin='start').std().lt(stdtol)
+    )
 
-    group = ((stationary_indicator != stationary_indicator.shift(1))
-             .cumsum()
-             .where(stationary_indicator))
-    stationary_len = (group.groupby(group, dropna=True, group_keys=False)
-                           .apply(lambda g: g.index[-1] - g.index[0]))
-    if len(stationary_len) > 0:
-        nonwear_len = stationary_len[stationary_len > pd.Timedelta(patience)]
-    else:
-        nonwear_len = pd.Series(dtype='timedelta64[ns]')  # empty series
+    segment_edges = (stationary_indicator != stationary_indicator.shift(1))
+    segment_edges.iloc[0] = True  # first edge is always True
+    segment_ids = segment_edges.cumsum()
+    stationary_segment_ids = segment_ids[stationary_indicator]
+    stationary_segment_lengths = (
+        stationary_segment_ids
+        .groupby(stationary_segment_ids)
+        .agg(
+            start_time=lambda x: x.index[0],
+            length=lambda x: x.index[-1] - x.index[0]
+        )
+        .set_index('start_time')
+        .squeeze(axis=1)
+        # dtype defaults to int64 when series is empty, so
+        # astype('timedelta64[ns]') makes sure it's always a timedelta,
+        # otherwise comparison with Timedelta(patience) below will fail
+        .astype('timedelta64[ns]')
+    )
+    nonwear_segment_lengths = stationary_segment_lengths[stationary_segment_lengths > pd.Timedelta(patience)]
 
-    nonwear_time = nonwear_len.sum().total_seconds()
-    wear_time, _ = get_wear_time(data.index.to_series())
-    info['WearTime(days)'] = (wear_time - nonwear_time) / (60 * 60 * 24)  # update wear time
-    info['NonwearTime(days)'] = nonwear_time / (60 * 60 * 24)
-    info['NumNonwearEpisodes'] = len(nonwear_len)
+    count_nonwear = len(nonwear_segment_lengths)
+    total_nonwear = nonwear_segment_lengths.sum().total_seconds()
+    total_wear = (
+        data.index.to_series().diff()
+        .pipe(lambda x: x[x < pd.Timedelta('1s')].sum())
+        .total_seconds()
+    ) - total_nonwear
 
-    # Flag nonwear
-    nonwear_indicator = group.isin(nonwear_len.index)
-    if drop:
-        data = data[~nonwear_indicator]
-    else:
-        data = data.mask(nonwear_indicator)
+    info['WearTime(days)'] = total_wear / (60 * 60 * 24)
+    info['NonwearTime(days)'] = total_nonwear / (60 * 60 * 24)
+    info['NumNonwearEpisodes'] = count_nonwear
+
+    # Flag nonwear segments
+    data = data.copy(deep=True)  # copy to avoid modifying original data
+    for start_time, length in nonwear_segment_lengths.items():
+        data.loc[start_time:start_time + length] = np.nan
 
     return data, info
 
 
-def calibrate_gravity(data, calib_cube=0.3, calib_min_samples=50, stationary_indicator=None):  # noqa: C901
+def calibrate_gravity(data, calib_cube=0.3, calib_min_samples=50, window='10s', stdtol=15 / 1000, chunksize=1_000_000):  # noqa: C901
     """
     Gravity calibration method of van Hees et al. 2014 (https://pubmed.ncbi.nlm.nih.gov/25103964/)
 
@@ -160,39 +248,50 @@ def calibrate_gravity(data, calib_cube=0.3, calib_min_samples=50, stationary_ind
         indicating stationary (low movement) periods. If None, it will be
         automatically inferred. Defaults to None.
     :type stationary_indicator: pandas.Series, optional
+    :param chunksize: Chunk size. Defaults to 1_000_000.
+    :type chunksize: int, optional
     :return: Processed data and processing info.
     :rtype: (pandas.DataFrame, dict)
     """
 
     info = {}
 
-    if stationary_indicator is None:
-        stationary_indicator = get_stationary_indicator(data)
+    stationary_indicator = (  # this is more memory friendly than of data[['x', 'y', 'z']].std()
+        data['x'].resample(window, origin='start').std().lt(stdtol)
+        & data['y'].resample(window, origin='start').std().lt(stdtol)
+        & data['z'].resample(window, origin='start').std().lt(stdtol)
+    )
 
-    # The paper uses 10sec averages instead of the raw ticks.
-    # This reduces computational cost. Also reduces influence of outliers.
-    stationary_data = (data[stationary_indicator]
-                       .resample('10s')
-                       .mean()
-                       .dropna())
-
-    hasT = 'temperature' in stationary_data
-
-    xyz = stationary_data[['x', 'y', 'z']].to_numpy()
+    xyz = (
+        data[['x', 'y', 'z']]
+        .resample(window, origin='start').mean()
+        [stationary_indicator]
+        .dropna()
+        .to_numpy()
+    )
     # Remove any nonzero vectors as they cause nan issues
     nonzero = np.linalg.norm(xyz, axis=1) > 1e-8
     xyz = xyz[nonzero]
+
+    hasT = 'temperature' in data
     if hasT:
-        T = stationary_data['temperature'].to_numpy()
+        T = (
+            data['temperature']
+            .resample(window, origin='start').mean()
+            [stationary_indicator]
+            .dropna()
+            .to_numpy()
+        )
         T = T[nonzero]
-    del stationary_data
+
+    del stationary_indicator
     del nonzero
 
     if len(xyz) < calib_min_samples:
         info['CalibOK'] = 0
         info['CalibErrorBefore(mg)'] = np.nan
         info['CalibErrorAfter(mg)'] = np.nan
-        warnings.warn(f"Skipping calibration: Insufficient stationary samples")
+        warnings.warn(f"Skipping calibration: Insufficient stationary samples: {len(xyz)} < {calib_min_samples}")
         return data, info
 
     intercept = np.array([0.0, 0.0, 0.0], dtype=xyz.dtype)
@@ -280,12 +379,50 @@ def calibrate_gravity(data, calib_cube=0.3, calib_min_samples=50, stationary_ind
         return data, info
 
     else:
-        data = data.copy()
-        data[['x', 'y', 'z']] = (best_intercept
-                                 + best_slope * data[['x', 'y', 'z']].to_numpy())
-        if hasT:
-            data[['x', 'y', 'z']] = (data[['x', 'y', 'z']]
-                                     + best_slopeT * (data['temperature'].to_numpy()[:, None]))
+
+        # # In-memory version
+        # data = data.copy()
+        # data[['x', 'y', 'z']] = (best_intercept
+        #                          + best_slope * data[['x', 'y', 'z']].to_numpy())
+        # if hasT:
+        #     data[['x', 'y', 'z']] = (data[['x', 'y', 'z']]
+        #                              + best_slopeT * (data['temperature'].to_numpy()[:, None]))
+
+        with tempfile.TemporaryDirectory() as tmpdir:
+            # We use TemporaryDirectory() + filename instead of NamedTemporaryFile()
+            # because we don't want to open the file just yet:
+            # https://stackoverflow.com/questions/26541416/generate-temporary-file-names-without-creating-actual-file-in-python
+            # and Windows doesn't allow opening a file twice:
+            # https://docs.python.org/3.9/library/tempfile.html#tempfile.NamedTemporaryFile
+
+            n = len(data)
+            mmap_fname = os.path.join(tmpdir, 'data.mmap')
+            data_mmap = mmap_like(data, mmap_fname, shape=(n,))
+
+            for i in range(0, n, chunksize):
+
+                # If last chunk, adjust chunksize
+                if i + chunksize > n:
+                    chunksize = n - i
+
+                chunk = data.iloc[i:i + chunksize]
+                chunk_xyz = chunk[['x', 'y', 'z']].to_numpy()
+                chunk_xyz = best_intercept + best_slope * chunk_xyz
+                if hasT:
+                    chunk_T = chunk['temperature'].to_numpy()
+                    chunk_xyz = chunk_xyz + best_slopeT * chunk_T[:, None]
+                chunk = chunk.copy(deep=True)  # copy to avoid modifying original data
+                chunk[['x', 'y', 'z']] = chunk_xyz
+
+                copy2mmap(chunk, data_mmap[i:i + chunksize])
+
+            del data
+
+            # We need to copy so that the mmap file can be trully deleted: 
+            # https://stackoverflow.com/questions/24178460/in-python-is-it-possible-to-overload-numpys-memmap-to-delete-itself-when-the-m
+            data = mmap2df(data_mmap, copy=True)
+
+            del data_mmap
 
         info['CalibOK'] = 1
         info['CalibNumIters'] = it + 1
@@ -421,5 +558,40 @@ def npy2df(data):
     t.name = 'time'
     columns = [c for c in data.dtype.names if c != 'time']
     data = pd.DataFrame({c: data[c] for c in columns}, index=t, copy=False)
+    return data
 
+
+def mmap_like(data, filename, mode='w+', shape=None):
+    dtype = np.dtype([
+        (data.index.name, data.index.dtype), 
+        *[(c, data[c].dtype) for c in data.columns]
+    ])
+    shape = shape or (len(data),)
+    data_mmap = np.memmap(filename, dtype=dtype, mode=mode, shape=shape)
+    return data_mmap
+
+
+def copy2mmap(data, data_mmap, flush=True):
+    """ Copy a pandas.DataFrame to a numpy.memmap. This operation is in-place.
+
+    :param data: A pandas.DataFrame of acceleration time-series.
+    :type data: pandas.DataFrame.
+    :param data_mmap: A numpy.memmap to copy data into.
+    :type data_mmap: numpy.memmap.
+    """
+    data_mmap[data.index.name] = data.index.to_numpy()
+    for c in data.columns:
+        data_mmap[c] = data[c].to_numpy()
+    if flush:
+        np.memmap.flush(data_mmap)
+    return
+
+
+def mmap2df(data_mmap, index_col='time', copy=True):
+    """ Convert a numpy structured array to pandas dataframe. """
+    columns = [c for c in data_mmap.dtype.names if c != index_col]
+    data = pd.DataFrame(
+        {c: np.asarray(data_mmap[c]) for c in columns}, copy=copy,
+        index=pd.Index(np.asarray(data_mmap[index_col]), name=index_col, copy=copy),
+    )
     return data
